const constants = require('.././constants.json');
const secrets = require('./secrets.json')
const fetch = require('node-fetch');
const alpha = 0.99
let log_channel = constants.topics.log_channel,
    metrics = {  }


let metricsDB = `http://${secrets.couchdb_username}:${secrets.couchdb_password}@${constants.couchdb_host}`
metricsDB = metricsDB + "/" + constants.metrics_db_name + "/"
let kafka = require('kafka-node'),
    Producer = kafka.Producer,
    client = new kafka.KafkaClient({
        kafkaHost: constants.network.external.kafka_host,
        autoConnect: true
    }),
    producer = new Producer(client)

/**
 * Function called to report metric data related to functions
 * @param {JSON} metric 
 */
function collectMetrics(metric) {
    /**
     * If metrics for a new function comes in, 
     * provision required structure for the function
     */
    if (!(metric.functionHash in metrics)) {
        metrics[metric.functionHash] = {
            shortterm: {
                coldstart: 0,
                coldstart_total_request: 0,
                warm_total_request: 0,
                warmstart: 0,
                worker_count: 0
            }
        }
    }
    if (metric.type === 'coldstart') {
        metrics[metric.functionHash].shortterm.coldstart += metric.value
        metrics[metric.functionHash].shortterm.coldstart_total_request += 1
    } else if (metric.type === 'warmstart') {
        metrics[metric.functionHash].shortterm.warmstart += metric.value
        metrics[metric.functionHash].shortterm.warm_total_request += 1
    } else if (metric.type === 'scale') {
        metrics[metric.functionHash].worker_count = metric.value
    }
    
}


/**
 * Run periodically to calculate average runtime metrics like coldstart and
 * warmstart latencies.
 * The module provides two granularities for metrics - shortterm and longterm
 * shortterm - realtime data at a granularity of 5s (set in dispatch_manager/lib.js)
 *             shortterm data is calculated using Simple Moving Average (SMA)
 * longterm - longterm data is held and averaged out over a period of time.
 *            longterm data is calculated using Expontential Moving Average (EMA)
 */
async function broadcastMetrics() {
    if (Object.keys(metrics).length !== 0) {
        for (let [functionHash, metricData] of Object.entries(metrics)) {
            if (metricData.shortterm.coldstart != 0 || metricData.shortterm.longterm != 0) {
                let { metric, revision } = await fetchData(functionHash, metricData)

                /**
                * Shortterm moving average
                */
                metric.shortterm.coldstart /= (metric.shortterm.coldstart_total_request != 0) ?
                    metric.shortterm.coldstart_total_request : 1
                metric.shortterm.warmstart /= (metric.shortterm.warm_total_request != 0) ?
                    metric.shortterm.warm_total_request : 1
                /**
                 * Longterm exponential moving average
                 */
                if (metric.shortterm.coldstart != 0)
                    metric.longterm.coldstart = (metric.longterm.coldstart != 0) ? metric.longterm.coldstart * alpha
                        + metric.shortterm.coldstart * (1 - alpha) : metric.shortterm.coldstart
                if (metric.shortterm.warmstart != 0)
                    metric.longterm.warmstart = (metric.longterm.warmstart != 0) ? metric.longterm.warmstart * alpha
                        + metric.shortterm.warmstart * (1 - alpha) : metric.shortterm.warmstart

                let payload = {
                    method: 'put',
                    body: JSON.stringify({
                        coldstart: metric.longterm.coldstart,
                        warmstart: metric.longterm.warmstart,
                        _rev: revision
                    }),
                    headers: { 'Content-Type': 'application/json' }
                }
                await fetch(metricsDB + functionHash, payload)
                metric.timestamp = Date.now()
            }
        }
        
        let log = [{
            topic: log_channel,
            messages: JSON.stringify({
                metrics
            }),
            partition: 0
        }]
        producer.send(log, () => { })
        
        for (let [functionHash, metric] of Object.entries(metrics)) {
            metric.shortterm = {
                coldstart: 0,
                coldstart_total_request: 0,
                warm_total_request: 0,
                warmstart: 0,
                worker_count: 0
            }
        }
    }
}

/**
 * Function to fetch the latest data from metric DB
 * @param {String} functionHash 
 * @param {JSON} metric 
 */
async function fetchData(functionHash, metric) {
    let revision
    let res = await fetch(metricsDB + functionHash)
    let json = await res.json()

    if (json.error === "not_found") {
        metric.longterm = {
            coldstart: 0,
            warmstart: 0
        }
    } else {

        metric.longterm = {
            coldstart: json.coldstart,
            warmstart: json.warmstart,
        }
        revision = json._rev
    }
    
    return {
        metric,
        revision
    }
}

module.exports = {
    collectMetrics, broadcastMetrics
}